; (x) herm1t@vx.netlux.org 2006-08-15
; *Description*
; RC4 generates a pseudorandom stream of bits (a "keystream") which, for
; encryption, is combined with the plaintext using XOR as with any Vernam
; cipher; decryption is performed the same way. To generate the keystream,
; the cipher makes use of a secret internal state which consists of two
; parts:
;   1. A permutation of all 256 possible bytes (denoted "S" below).
;   2. Two 8-bit index-pointers (denoted "i" and "j").
; The permutation is initialised with a variable length key, typically
; between 40 and 256 bits, using the key-scheduling algorithm (KSA). Once
; this has been completed, the stream of bits is generated using the
; pseudo-random generation algorithm (PRGA).
;   The key-scheduling algorithm (KSA)
; The key-scheduling algorithm is used to initialise the permutation in
; the array "S". "keylength" is defined as the number of bytes in the key
; and can be in the range 1 <= keylength <= 256, typically between 5 and 16,
; corresponding to a key length of 40128 bits. First, the array "S" is
; initialised to the identity permutation. S is then processed for 256
; iterations in a similar way to the main PRGA algorithm, but also mixes in
; bytes of the key at the same time.
;   The pseudo-random generation algorithm (PRGA)
; For as many iterations as are needed, the PRGA modifies the state and
; outputs a byte of the keystream. In each iteration, the PRGA increments i,
; adds the value of S pointed to by i to j, exchanges the values of S[i] and
; S[j], and then outputs the value of S at the location S[i] + S[j] (modulo
; 256). Each value of S is swapped at least once every 256 iterations.

		BITS	32
		CPU	386
		global	rc4_expandkey,rc4_cipher

;void rc4_expandkey(unsigned char *S, unsigned char *key, int len)
;{
;	register int	i, j;
;	for (i = 0; i < 256; i++)
;		S[i] = i;
;	j = 0;
;	for (i = 0; i < 256; i++) {
;		j = (j + S[i] + key[i % len]) % 256;
;		SWAP(S[i], S[j]);
;	}
;}
rc4_expandkey:	pusha
		mov	ebp, [esp + 36]	; key
		mov	esi, [esp + 40]	; user key
		mov	edi, [esp + 44]	; user key length
		xor	eax, eax
	.l0:	mov	[ebp + eax], al
		inc	eax
		or	ah, ah
		jz	.l0
		xor	eax, eax
		xor	ebx, ebx
	.l1:	push	eax
		xor	edx, edx
		div	edi
		pop	eax
		movzx	edx, byte [esi + edx]
		add	bl, dl
		mov	dl, [ebp + eax]
		add	bl, dl
		xchg	dl, [ebp + ebx]
		xchg	dl, [ebp + eax]
		inc	eax
		or	ah, ah
		jz	.l1
		popa
		ret

;void rc4_cipher(unsigned char *src, unsigned char *S, int len)
;{
;	register int	i, j;
;	i = 0;
;	j = 0;
;	while (len-- > 0) {
;		i = (i + 1) % 256;
;		j = (j + S[i]) % 256;
;		SWAP(S[i], S[j]);
;		*src++ ^= S[(S[i] + S[j]) % 256];
;	}
;}
rc4_cipher:	pusha
		mov	esi, [esp + 36]	; src
		mov	edi, esi
		mov	ebp, [esp + 40]	; key
		mov	ecx, [esp + 44]	; len
		xor	eax, eax
		xor	ebx, ebx
	.while:	inc	al
		movzx	edx, byte [ebp + eax]
		add	bl, dl
		xchg	dl, [ebp + ebx]
		xchg	dl, [ebp + eax]
		add	dl, [ebp + eax]
		mov	dl, [ebp + edx]
		push	eax
		lodsb
		xor	eax, edx
		stosb
		pop	eax
		loop	.while
		popa
		ret
